import sys
import os
import string
import filecmp
import shutil

forge = sys.modules["forge"]

def setup(module):
	forge.extPath = os.path.join(forge.rootPath,'ext')
	forge.includemap['ext'] = forge.extPath
	suppress = {}

	ext_string = os.getenv('FE_EXT')
	if ext_string == None or ext_string == "":
		ext_string = forge.extensions

	if ext_string != None:
		if ext_string != "":
			extensions = string.split(ext_string, ':')
			for extension in extensions:
				forge.module_list.append(extension)
	else:
		for name in os.listdir(forge.extPath):
			dir_path = os.path.join(forge.extPath, name)
			if os.path.isdir(dir_path):
				bld_py_path = os.path.join(dir_path, 'bld.py')
				if os.path.exists(bld_py_path):
					forge.module_list.append(name)
	ext_no_string = os.getenv('FE_NO_EXT')
	if ext_no_string != None:
		no_extensions = string.split(ext_no_string, ':')
		for extension in no_extensions:
			try:
				suppress[extension] = 1
			except:
				pass

	manifests = 'ext/autoload/manifests.cc'
	manifests_new = 'ext/autoload/manifests_new.cc'
	with open(manifests_new, 'w') as outfile:
		outfile.write("\t// this file is auto-generated\n\n")

	forge.cprint(forge.CYAN,1, 'Extensions: %d' % len(forge.module_list))
	forge.module_list.sort()
	for i in forge.module_list:
		forge.color_on(0, forge.CYAN)
		sys.stdout.write('(')
		forge.color_off()
		if suppress.has_key(i):
			forge.color_on(1, forge.YELLOW)
			sys.stdout.write('%s' % (i))
		else:
			child = module.Module(i)
			clr = forge.CYAN

			if not child:
				clr = forge.RED
				forge.color_on(0, forge.RED)
				sys.stdout.write('fail %s ' % (i))
				forge.color_off()

			if child.ok != None:
				clr = forge.RED
				forge.color_on(0, forge.RED)
				sys.stdout.write('fail:%s ' % (child.ok))
				forge.color_off()

			forge.color_on(1, clr)
			sys.stdout.write('%s'% (i))
			forge.color_off()

			if child.hasDLL:
				forge.color_on(0, forge.CYAN)
				sys.stdout.write(' dl')
				forge.color_off()

				manifest = 'ext/'+i+'/manifest.cc'
				if os.path.exists(manifest):
					forge.color_on(0, forge.MAGENTA)
					sys.stdout.write(' auto')
					forge.color_off()

					with open(manifests_new, 'a') as outfile:
						with open(manifest) as infile:
							outfile.write('\t// '+i+'\n')
							outfile.write(infile.read())

			if child.hasLib:
				forge.color_on(0, forge.CYAN)
				sys.stdout.write(' lib')
				forge.color_off()

			if child.hasExe:
				forge.color_on(0, forge.CYAN)
				sys.stdout.write(' exe')
				forge.color_off()

		forge.color_on(0, forge.CYAN)
		sys.stdout.write(') ')
		forge.color_off()

	# don't trigger rebuild unless changed
	if not filecmp.cmp(manifests_new,manifests):
		shutil.copyfile(manifests_new,manifests)
	os.remove(manifests_new)

	forge.color_off()
	sys.stdout.write('\n')

	forge.help.append("""\
Free Electron extensions specific help
Environment Variables:
FE_EXT            - build only specific extensions
FE_NO_EXT         - avoid building specific extensions
""")

